import java.util.*;

class ListaDI extends AbstractSequentialList {
  Nod cap, coada; // referinte la nodurile sentinela
  int lungime; // numarul de elemente din lista (excluzand sentinelele)

  class Nod {
    Nod precedent, urmator;
    Object element;

    Nod(Object obj) { // constructorul nodului
      element=obj;
      precedent=urmator=null;
    }
  }

  class Iter implements ListIterator {
    Nod cursor;
    int indice;

    public boolean hasNext() {
      if(cursor.urmator==null || cursor.urmator.urmator==null)
              return false;
      return true;
    }

    public Object next() throws NoSuchElementException {
      if(!hasNext()) throw new NoSuchElementException();
      cursor=cursor.urmator;
      indice++;
      return cursor.element;
    }

    public boolean hasPrevious() {
      if(cursor.precedent==null) return false;
      return true;
    }

    public Object previous() throws NoSuchElementException {
      if(!hasPrevious()) throw new NoSuchElementException();
      Object elementPrecedent=cursor.element;
      cursor=cursor.precedent;
      indice--;
      return elementPrecedent;
    }

    public int nextIndex() {
      return indice+1;
    }

    public int previousIndex() {
      return indice;
    }

    public void remove() throws IllegalStateException {
      if(lungime==0) throw new IllegalStateException();
      cursor.precedent.urmator=cursor.urmator;
      cursor.urmator.precedent=cursor.precedent;
      lungime--;
    }

    public void set(Object obj) throws IllegalStateException { 
      if(cursor.precedent==null || cursor.urmator==null)
              throw new IllegalStateException();
      cursor.element=obj;
    }

    public void add(Object obj) {
      Nod nodNou=new Nod(obj);
      nodNou.precedent=cursor;
      nodNou.urmator=cursor.urmator;
      cursor.urmator.precedent=nodNou;
      cursor.urmator=nodNou;
      cursor=nodNou;
      indice++;
      lungime++; 
    }
  }  

  ListaDI() {
    cap=new Nod(null);   // nod sentinela
    coada=new Nod(null); // nod sentinela
    cap.urmator=coada; // legatura directa intre sentinele
    coada.precedent=cap;
    lungime=0;
  }

  public ListIterator listIterator(int index) {
    Iter iterator=new Iter();
    if (index<0) index=0;
    else if(index>lungime) index=lungime;
    iterator.cursor=cap;
    iterator.indice=-1;
    for(int i=0; i<index; i++) {
         iterator.cursor=iterator.cursor.urmator;
         iterator.indice++;
    }
    return iterator;
  }

  public int size() {
    return lungime;
  } 
}

class TestListaDI {
  public static void main(String args[]) {
    ListaDI lista=new ListaDI();
    String[] ts={"A", "B", "C", "D"};
    System.out.println("Lista initiala "+lista);
    for(int i=0; i<ts.length; i++) {
      lista.add(ts[i]);
    }
    System.out.println("Dupa adaugarea sirurilor din ts: "+lista);
    System.out.println("Lungimea listei: "+lista.size());
    ListIterator iter=lista.listIterator();
    while(iter.hasNext()) {
       Object obj=iter.next();
       System.out.print(obj+" ");
    }
    System.out.println();
    System.out.println("In ordine inversa");
    while(iter.hasPrevious()) {
      Object obj=iter.previous();
      System.out.print(obj+" ");
    }
    System.out.println();
    lista.remove("C");
    System.out.println("Dupa eliminarea elementului C: "+lista);
    System.out.println("Lungimea listei: "+lista.size());
    String[] ts1={"A1", "A2", "A3", "B1"}, ts2={"A2", "B"};
    List lista1=Arrays.asList(ts1), lista2=Arrays.asList(ts2);
    lista.addAll(lista1);
    System.out.println("S-a adaugat lista1: "+lista);
    System.out.println("Contine obiectul A1: "+
      lista.contains("A1"));
    System.out.println("Ultimul indice al lui A: "+
      lista.lastIndexOf("A"));
    System.out.println("Contine lista2: "+lista.containsAll(lista2));
    lista.removeAll(lista2);
    System.out.println("Dupa extragere: "+lista);
 
  }
}
