/* Clasa de grafuri orientate, in care fiecare varf al grafului 
   este o instanta a clasei interioare Varf si contine o eticheta 
   (din clasa String) si o informatie atasata.
   Fiecare arc este o instanta a clasei interioare Arc si contine
   o referinta la varful de destinatie si o informatie atasata (de
   exemplu o eticheta sau o pondere).

   Varfurile grafului sunt continute in campul varfuri, care este o 
   instanta a clasei ArrayList. Punerea varfurilor in aceasta lista
   se face prin metoda puneVarf. Aceasta metoda le pune in ordinea
   etichetelor, astfel ca lista varfuri este continuu sortata. Cautarea 
   varfurilor in lista se face prin algoritmul cautarii binare.
*/

import java.util.*;
import java.io.*;

public class Graf implements Serializable {
  protected ArrayList varfuri;

  public Graf() {
    varfuri=new ArrayList();
  }

  /* Cautare binara recursiva in intervalul [i1, i2]. 
     Valoarea intoarsa este indicele pozitiei pe care ar trebui sa se
     gaseasca elementul cautat (sau unde ar putea fi inserat)
  */
  protected int cautaLocVarf(Varf v, int i1, int i2) {
    if(i1>i2) return -2;
    int im=(i1+i2)/2, rezcomp=v.compareTo(varfuri.get(im));
    if(rezcomp==0) return im;
    if(rezcomp<0) 
     if(im==i1) return im;
     else return cautaLocVarf(v, i1, im-1);
    if(im==i2) return im+1;
    return cautaLocVarf(v, im+1, i2);
  }
    
  /* Cautare binara a unui varf in lista varfuri */
  protected int indiceVarf(String eticheta) {
    Varf varfCautat=new Varf(eticheta, null);
    int nrVarfuri=varfuri.size();
    int indice=cautaLocVarf(varfCautat, 0, nrVarfuri-1);
    if(indice<0) return indice;
    if(indice>=varfuri.size()) return -1;
    if(varfuri.get(indice).equals(varfCautat)) return indice;
    return -1;
  }

  public boolean contineVarf(String eticheta) {
    int indice=indiceVarf(eticheta);
    if(indice<0) return false;
    return true;
  }

  public Object infoVarf(String eticheta) {
    int indice=indiceVarf(eticheta);
    if(indice<0) return null;
    return ((Varf)varfuri.get(indice)).info;
  }

  public boolean contineArc(String etbaza, String etvarf) {
    int indiceBaza=indiceVarf(etbaza);
    if(indiceBaza<0) return false;
    int indiceVarf=indiceVarf(etvarf);
    if(indiceVarf<0) return false;
    Varf baza=(Varf)varfuri.get(indiceBaza);
    Varf varf=(Varf)varfuri.get(indiceVarf);
    return baza.arce.contains(new Arc(varf, null));
  }

  public Object infoArc(String etbaza, String etvarf) {
    int indiceBaza=indiceVarf(etbaza);
    if(indiceBaza<0) return null;
    int indiceVarf=indiceVarf(etvarf);
    if(indiceVarf<0) return null;
    Varf baza=(Varf)varfuri.get(indiceBaza);
    Varf varf=(Varf)varfuri.get(indiceVarf);
    int indiceArc=baza.arce.indexOf(new Arc(varf, null));
    if(indiceArc<0) return null;
    return ((Arc)baza.arce.get(indiceArc)).info;
  }


  /* Punerea unui nou varf in lista, respectand sortarea dupa
     etichete
  */
  public boolean puneVarf(String eticheta, Object info) {
    if(eticheta==null) return false;
    int nrVarfuri=varfuri.size();
    Varf varfNou=new Varf(eticheta, info);
    if(nrVarfuri==0) {
      varfuri.add(varfNou);
      return true;
    }
    int indice=cautaLocVarf(varfNou, 0, nrVarfuri-1);
    if(indice<0) return false;
    if(indice<nrVarfuri && varfuri.get(indice).equals(varfNou))
      return false;
    varfuri.add(indice, varfNou);
    return true;
  }

  /* Punerea unui nou arc in lista de arce a varfului sau de baza.
     Etichetele varfurilor de origine si de destinatie sunt, respectiv,
     etbaza si etvarf.
  */
  public boolean puneArc(String etbaza, String etvarf, Object info) {
    if(etbaza==null || etvarf==null) return false;
    int index;
    index=indiceVarf(etbaza);
    if(index<0) return false;
    Varf vbaza=(Varf)varfuri.get(index);
    index=indiceVarf(etvarf);
    Varf vvarf=(Varf)varfuri.get(index);
    Arc arcNou=new Arc(vvarf, info);
    if(vbaza.arce.contains(arcNou)) return false;
    vbaza.arce.add(arcNou);
    return true;
  }

  /* Pune in varful cu eticheta "etichetaVarf" informatia "info" si
     intoarce vechea informatie din varful respectiv
  */
  public Object puneInfoVarf(String etichetaVarf, Object info) {
    int indiceVarf=indiceVarf(etichetaVarf);
    if(indiceVarf<0) return null;
    Varf varf=(Varf)varfuri.get(indiceVarf);
    Object infoVechi=varf.info;
    varf.info=info;
    return infoVechi;
  }

  /* Se pune o noua informatie in arcul de la varful cu eticheta
     "etbaza" la varful cu eticheta "etvarf" si se intoarce
      informatia anterioara din arcul respectiv
  */
  public Object puneInfoArc(String etbaza, String etvarf, Object info){
    int indiceBaza=indiceVarf(etbaza);
    if(indiceBaza<0) return null;
    int indiceVarf=indiceVarf(etvarf);
    if(indiceVarf<0) return null;
    Varf baza=(Varf)varfuri.get(indiceBaza);
    Varf varf=(Varf)varfuri.get(indiceVarf);
    int indiceArc=baza.arce.indexOf(new Arc(varf, null));
    if(indiceArc<0) return null;
    Arc arc=(Arc)baza.arce.get(indiceArc);
    Object infoVechi=arc.info;
    arc.info=info;
    return infoVechi;
  }

  /* Se elimina arcul dintre varfurile cu etichetele etbaza si etvarf
  */
  public boolean eliminaArc(String etbaza, String etvarf) {
    if(etbaza==null || etvarf==null) return false;
    int index;
    index=indiceVarf(etbaza);
    if(index<0) return false;
    Varf vbaza=(Varf)varfuri.get(index);
    index=indiceVarf(etvarf);
    if(index<0) return false;
    Varf vvarf=(Varf)varfuri.get(index);
    return vbaza.arce.remove(new Arc(vvarf, null));
  }

  /* Se elimina varful cu eticheta etVarf */
  public boolean eliminaVarf(String etvarf) {
    if(etvarf==null) return false;
    int index=indiceVarf(etvarf);
    if(index<0) return false;
    Varf varfEliminat=(Varf)varfuri.get(index);
    /* Se elimina de la toate varfurile arcele care duc catre
       varful cu eticheta etvarf
    */
    Iterator iter=varfuri.iterator();
    Varf varf;
    while(iter.hasNext()) {
      varf=(Varf)iter.next();
      varf.arce.remove(new Arc(varfEliminat, null));
    }
    /* Se elimina insusi varful cu eticheta etvarf */
    varfuri.remove(varfEliminat);
    return true;
  }

  /* Intoarce un iterator pentru explorarea in latime a grafului,
     pornind de la varful cu eticheta etvarf
  */
  public Iterator iteratorLatime(String etichetaVarf) {
    if(etichetaVarf==null) return null;
    int index=indiceVarf(etichetaVarf);
    if(index<0) return null;
    Varf varf=(Varf)varfuri.get(index);
    return new IteratorLatime(varf);
  }

  /* Intoarce un iterator pentru explorarea in adancime a grafului,
     pornind de la varful cu eticheta etvarf
  */
  public Iterator iteratorAdancime(String etichetaVarf) {
    if(etichetaVarf==null) return null;
    int index=indiceVarf(etichetaVarf);
    if(index<0) return null;
    Varf varf=(Varf)varfuri.get(index);
    return new IteratorAdancime(varf);
  }

  /* Intoarce lista varfurilor pentru explorare in latime */
  public List explorareLatime(String etichetaVarfPornire) {
    Iterator iter=iteratorLatime(etichetaVarfPornire);
    if(iter==null) return null;
    List lista=new ArrayList();
    while(iter.hasNext()) lista.add(iter.next());
    return lista;
  }

  /* Intoarce lista varfurilor pentru explorare in adancime */
  public List explorareAdancime(String etichetaVarfPornire) {
    Iterator iter=iteratorAdancime(etichetaVarfPornire);
    if(iter==null) return null;
    List lista=new ArrayList();
    while(iter.hasNext()) lista.add(iter.next());
    return lista;
  }


  public String toString() {
    return varfuri.toString();
  } 

  /* Clasa interioara a varfurilor grafului */
  public class Varf implements Comparable, Serializable {
    protected String eticheta;
    protected Object info;
    protected ArrayList arce;

    public Varf(String eticheta, Object info) {
      this.eticheta=eticheta;
      this.info=info;
      arce=new ArrayList();
    }

    public String eticheta() {
      return eticheta;
    }

    public Object informatie() {
      return info;
    }

    /* Doua varfuri sunt echivalente daca au aceeasi eticheta */
    public boolean equals(Object obj) {
      if(obj==null) return false;
      if(obj==this) return true;
      return eticheta.equals(((Varf)obj).eticheta);
    }

    /* Ordonarea varfurilor se face dupa eticheta */
    public int compareTo(Object obj) {
      return eticheta.compareTo(((Varf)obj).eticheta);
    }

    public String toString() {
      return eticheta+"="+info+arce;
    }
  }

  /* Clasa interioara a arcelor grafului */
  public class Arc implements Serializable {
    protected Varf varf;
    protected Object info;

    public Arc(Varf destinatie, Object informatie) {
      varf=destinatie;
      info=informatie;
    }

    public Object informatie() {
      return info;
    }

    /* Doua arce care au aceeasi origine sunt echivalente daca au
       si acelasi varf
    */
    public boolean equals(Object obj) {
      return varf.equals(((Arc)obj).varf);
    }

    public String toString() {
      return varf.eticheta+"="+info;
    }
  }

  /* Clasa interioara a iteratorilor pentru explorarea in latime */
  protected class IteratorLatime implements Iterator {
    LinkedList coada;
    TreeSet parcurse;
    
    IteratorLatime(Varf varfPornire) {
      coada=new LinkedList();
      parcurse=new TreeSet();
      coada.add(varfPornire);
      parcurse.add(varfPornire);
    }

    public boolean hasNext() {
      if(coada.size()>0) return true;
      return false;
    }

    public Object next() {
      if(coada.size()==0) return null;
      Varf varfCurent=(Varf)coada.removeFirst();
      int nrArce=varfCurent.arce.size();
      Varf varf;
      for(int i=0; i<nrArce; i++)  {
        varf=((Arc)varfCurent.arce.get(i)).varf;
        if(!parcurse.contains(varf)) {
          coada.addLast(varf);
          parcurse.add(varf);
        }
      }
      return varfCurent;
    }

    public void remove() {
    }
  }

  /* Clasa interioara a iteratorilor pentru explorarea in adancime */
  protected class IteratorAdancime implements Iterator {
    Stack stiva;
    TreeSet parcurse;
    boolean pusStiva;
    
    IteratorAdancime(Varf varfPornire) {
      stiva=new Stack();
      parcurse=new TreeSet();
      stiva.push(new ElementStiva(varfPornire));
      parcurse.add(varfPornire);
      pusStiva=true;
    }

    public boolean hasNext() {
      if(pusStiva) return true;
      /* Se fac operatiile necesare ca varful urmator sa ajunga
         in varful stivei
      */
      ElementStiva es;
      Varf varfUrmator=null;
      boolean gasit=false;
      while(!gasit) {
        if(stiva.size()==0) return false;
        es=(ElementStiva)stiva.peek();
        if(es.iterator.hasNext()){
          varfUrmator=((Arc)es.iterator.next()).varf;
          if(!parcurse.contains(varfUrmator)) {
            stiva.push(new ElementStiva(varfUrmator));
            parcurse.add(varfUrmator);
            pusStiva=true;
            gasit=true;
          }
        }
        else {
         stiva.pop(); // se scoate varful cu arcele epuizate
        }
      }
      return gasit; 
    }

    public Object next() {
      if(!hasNext()) return null;
      if(pusStiva) {
        pusStiva=false;
        return ((ElementStiva)stiva.peek()).varf;
      }
      return null;
    }

    public void remove() {
    }
  }

  /* Elementele stivei folosite la explorarea in adancime */
  protected class ElementStiva {
    Varf varf;
    Iterator iterator;

    ElementStiva(Varf v) {
      varf=v;
      iterator=v.arce.iterator();
    }

    public String toString() {
      return varf.eticheta;
    }
  }
}
